@@ -186,7 +186,7 @@ def restore_original_settings():
 def set_default_settings():
     global ENGINE
     global TABSQLITEDB
-    ENGINE.set_input_mode(mode=1)
+    ENGINE.set_input_mode(input_mode=1)
     chinese_mode = 4
     language_filter = TABSQLITEDB.ime_properties.get('language_filter')
     if language_filter in ('cm0', 'cm1', 'cm2', 'cm3', 'cm4'):
@@ -412,6 +412,48 @@ def tear_down():
     assert IBus.PropList is IBUS_PROP_LIST
     assert IBus.PropList is not MockPropList
 
+def require_serial_number(required_serial_number):
+    serial_number = int(TABSQLITEDB.ime_properties.get('serial_number'))
+    return serial_number >= required_serial_number
+
+class ErbiQsTestCase(unittest.TestCase):
+    def setUp(self):
+        engine_name = 'erbi-qs'
+        if not set_up(engine_name):
+            self.skipTest('Could not setup “%s”, skipping test.' % engine_name)
+
+    def tearDown(self):
+        tear_down()
+
+    def test_dummy(self):
+        self.assertEqual(True, True)
+
+    def test_get_goucima(self):
+        if not require_serial_number(20220111):
+            self.skipTest('serial_number too small')
+        self.assertEqual(
+            'sxr.',
+            TABSQLITEDB.get_goucima('松'))
+        self.assertEqual(
+            'thr',
+            TABSQLITEDB.get_goucima('天'))
+        self.assertEqual(
+            'xjv',
+            TABSQLITEDB.get_goucima('下'))
+        self.assertEqual(
+            'dkv',
+            TABSQLITEDB.get_goucima('大'))
+        self.assertEqual(
+            'sjnb',
+            TABSQLITEDB.get_goucima('事'))
+
+    def test_parse_phrase(self):
+        if not require_serial_number(20220111):
+            self.skipTest('serial_number too small')
+        self.assertEqual(
+            'txds',
+            TABSQLITEDB.parse_phrase('天下大事'))
+
 class WubiJidian86TestCase(unittest.TestCase):
     def setUp(self):
         engine_name = 'wubi-jidian86'
@@ -424,6 +466,11 @@ class WubiJidian86TestCase(unittest.TestCase):
     def test_dummy(self):
         self.assertEqual(True, True)
 
+    def test_parse_phrase(self):
+        self.assertEqual(
+            'ggdg',
+            TABSQLITEDB.parse_phrase('天下大事'))
+
     def test_single_char_commit_with_space(self):
         ENGINE._do_process_key_event(IBus.KEY_a, 0, 0)
         ENGINE._do_process_key_event(IBus.KEY_space, 0, 0)
@@ -1069,81 +1116,91 @@ class WubiJidian86TestCase(unittest.TestCase):
         ENGINE.set_chinese_mode(
             mode=0, update_gsettings=False) # show simplified Chinese only
         ENGINE._do_process_key_event(IBus.KEY_c, 0, 0)
-        self.assertEqual(ENGINE._lookup_table.mock_candidates,
-                         ['以  418261033 0',
-                          '能 ex 1820000000 0',
-                          '能 exx 1820000000 0',
-                          '对 fy 1200000000 0',
-                          '又 cc 729000000 0',
-                          '又 ccc 729000000 0',
-                          '通 ep 521000000 0',
-                          '通 epk 521000000 0',
-                          '台 kf 486000000 0',
-                          '难忘 wyn 404000000 0'])
+        ENGINE._do_process_key_event(IBus.KEY_g, 0, 0)
+        self.assertEqual(['对不起 fh 44400000 0',
+                          '能不能 ce 21400000 0',
+                          '又不是 jg 19000000 0',
+                          '马来西亚 sg 14100000 0',
+                          '参与者 ft 6390000 0',
+                          '骊 m 5420000 0',
+                          '骊 my 5420000 0',
+                          '邓丽君 vt 4160000 0',
+                          '对不对 cf 2720000 0',
+                          '对不住 wy 2110000 0'],
+                         ENGINE._lookup_table.mock_candidates,)
+        ENGINE._do_process_key_event(IBus.KEY_BackSpace, 0, 0)
         ENGINE._do_process_key_event(IBus.KEY_BackSpace, 0, 0)
         self.assertEqual(ENGINE._lookup_table.mock_candidates, [])
         ENGINE.set_chinese_mode(
             mode=1, update_gsettings=False) # show traditional Chinese only
         ENGINE._do_process_key_event(IBus.KEY_c, 0, 0)
-        self.assertEqual(ENGINE._lookup_table.mock_candidates,
-                         ['以  418261033 0',
-                          '能 ex 1820000000 0',
-                          '能 exx 1820000000 0',
-                          '又 cc 729000000 0',
-                          '又 ccc 729000000 0',
-                          '通 ep 521000000 0',
-                          '通 epk 521000000 0',
-                          '台 kf 486000000 0',
-                          '能 e 306980312 0',
-                          '能力 elt 274000000 0'])
+        ENGINE._do_process_key_event(IBus.KEY_g, 0, 0)
+        self.assertEqual(['驪  0 0',
+                          '馬 gy 99400000 0',
+                          '馬 hy 99400000 0',
+                          '对不起 fh 44400000 0',
+                          '能不能 ce 21400000 0',
+                          '又不是 jg 19000000 0',
+                          '对不对 cf 2720000 0',
+                          '对不住 wy 2110000 0',
+                          '难不成 dn 1990000 0',
+                          '巴不得 tj 1420000 0'],
+                         ENGINE._lookup_table.mock_candidates)
+        ENGINE._do_process_key_event(IBus.KEY_BackSpace, 0, 0)
         ENGINE._do_process_key_event(IBus.KEY_BackSpace, 0, 0)
         self.assertEqual(ENGINE._lookup_table.mock_candidates, [])
         ENGINE.set_chinese_mode(
             mode=2, update_gsettings=False) # show simplified Chinese first
         ENGINE._do_process_key_event(IBus.KEY_c, 0, 0)
-        self.assertEqual(ENGINE._lookup_table.mock_candidates,
-                         ['以  418261033 0',
-                          '能 ex 1820000000 0',
-                          '能 exx 1820000000 0',
-                          '对 fy 1200000000 0',
-                          '又 cc 729000000 0',
-                          '又 ccc 729000000 0',
-                          '通 ep 521000000 0',
-                          '通 epk 521000000 0',
-                          '台 kf 486000000 0',
-                          '难忘 wyn 404000000 0'])
+        ENGINE._do_process_key_event(IBus.KEY_g, 0, 0)
+        self.assertEqual(['驪  0 0',
+                          '对不起 fh 44400000 0',
+                          '能不能 ce 21400000 0',
+                          '又不是 jg 19000000 0',
+                          '马来西亚 sg 14100000 0',
+                          '参与者 ft 6390000 0',
+                          '骊 m 5420000 0',
+                          '骊 my 5420000 0',
+                          '邓丽君 vt 4160000 0',
+                          '对不对 cf 2720000 0'],
+                         ENGINE._lookup_table.mock_candidates)
+        ENGINE._do_process_key_event(IBus.KEY_BackSpace, 0, 0)
         ENGINE._do_process_key_event(IBus.KEY_BackSpace, 0, 0)
         self.assertEqual(ENGINE._lookup_table.mock_candidates, [])
         ENGINE.set_chinese_mode(
             mode=3, update_gsettings=False) # show traditional Chinese first
         ENGINE._do_process_key_event(IBus.KEY_c, 0, 0)
-        self.assertEqual(ENGINE._lookup_table.mock_candidates,
-                         ['以  418261033 0',
-                          '能 ex 1820000000 0',
-                          '能 exx 1820000000 0',
-                          '又 cc 729000000 0',
-                          '又 ccc 729000000 0',
-                          '通 ep 521000000 0',
-                          '通 epk 521000000 0',
-                          '台 kf 486000000 0',
-                          '能 e 306980312 0',
-                          '能力 elt 274000000 0'])
+        ENGINE._do_process_key_event(IBus.KEY_g, 0, 0)
+        self.assertEqual(['驪  0 0',
+                          '馬 gy 99400000 0',
+                          '馬 hy 99400000 0',
+                          '对不起 fh 44400000 0',
+                          '能不能 ce 21400000 0',
+                          '又不是 jg 19000000 0',
+                          '对不对 cf 2720000 0',
+                          '对不住 wy 2110000 0',
+                          '难不成 dn 1990000 0',
+                          '巴不得 tj 1420000 0'],
+                         ENGINE._lookup_table.mock_candidates)
+        ENGINE._do_process_key_event(IBus.KEY_BackSpace, 0, 0)
         ENGINE._do_process_key_event(IBus.KEY_BackSpace, 0, 0)
         self.assertEqual(ENGINE._lookup_table.mock_candidates, [])
         ENGINE.set_chinese_mode(
             mode=4, update_gsettings=False) # show all characters
         ENGINE._do_process_key_event(IBus.KEY_c, 0, 0)
-        self.assertEqual(ENGINE._lookup_table.mock_candidates,
-                         ['以  418261033 0',
-                          '能 ex 1820000000 0',
-                          '能 exx 1820000000 0',
-                          '对 fy 1200000000 0',
-                          '又 cc 729000000 0',
-                          '又 ccc 729000000 0',
-                          '通 ep 521000000 0',
-                          '通 epk 521000000 0',
-                          '台 kf 486000000 0',
-                          '难忘 wyn 404000000 0'])
+        ENGINE._do_process_key_event(IBus.KEY_g, 0, 0)
+        self.assertEqual(['驪  0 0',
+                          '馬 gy 99400000 0',
+                          '馬 hy 99400000 0',
+                          '对不起 fh 44400000 0',
+                          '能不能 ce 21400000 0',
+                          '又不是 jg 19000000 0',
+                          '马来西亚 sg 14100000 0',
+                          '参与者 ft 6390000 0',
+                          '骊 m 5420000 0',
+                          '骊 my 5420000 0'],
+                         ENGINE._lookup_table.mock_candidates)
+        ENGINE._do_process_key_event(IBus.KEY_BackSpace, 0, 0)
         ENGINE._do_process_key_event(IBus.KEY_BackSpace, 0, 0)
         self.assertEqual(ENGINE._lookup_table.mock_candidates, [])
 
@@ -1277,9 +1334,11 @@ class Cangjie5TestCase(unittest.TestCase):
         self.assertEqual(ENGINE.mock_committed_text, '日')
 
     def test_type_one_char_and_check_auxiliary(self):
+        if not require_serial_number(20220111):
+            self.skipTest('serial_number too small')
         ENGINE._do_process_key_event(IBus.KEY_d, 0, 0)
         self.assertEqual(ENGINE.mock_preedit_text, '木')
-        self.assertEqual(ENGINE._lookup_table.mock_candidates[8],
+        self.assertEqual(ENGINE._lookup_table.mock_candidates[6],
                          '林 木 1000 0')
         ENGINE._do_process_key_event(IBus.KEY_v, 0, 0)
         ENGINE._do_process_key_event(IBus.KEY_i, 0, 0)
@@ -1292,41 +1351,45 @@ class Cangjie5TestCase(unittest.TestCase):
         self.assertEqual(ENGINE.mock_committed_text, '機')
 
     def test_dynamic_adjust(self):
+        if not require_serial_number(20220111):
+            self.skipTest('serial_number too small')
         ENGINE.set_dynamic_adjust(False)
         ENGINE._do_process_key_event(IBus.KEY_i, 0, 0)
         ENGINE._do_process_key_event(IBus.KEY_r, 0, 0)
         ENGINE._do_process_key_event(IBus.KEY_p, 0, 0)
-        self.assertEqual('怠', ENGINE.mock_preedit_text)
+        self.assertEqual('感', ENGINE.mock_preedit_text)
         # Big5 code of 怠 is ABE5 (see
         # https://www.unicode.org/cgi-bin/GetUnihanData.pl?codepoint=%E6%84%9F)
         # Big5 code of 感 is B750 (see
         # https://www.unicode.org/cgi-bin/GetUnihanData.pl?codepoint=%E6%80%A0)
+        #
+        # In the improved cangjie5 table, 感 has higher priority now.
         self.assertEqual(
-            ['怠  1000 0',
-             '感  1000 0'],
+            ['感  1000 0',
+             '怠  900 0'],
             ENGINE._lookup_table.mock_candidates)
         ENGINE._do_process_key_event(IBus.KEY_2, 0, 0)
         self.assertEqual(
             [],
             ENGINE._lookup_table.mock_candidates)
         self.assertEqual('', ENGINE.mock_preedit_text)
-        self.assertEqual('感', ENGINE.mock_committed_text)
+        self.assertEqual('怠', ENGINE.mock_committed_text)
         # repeat the same input, result should be the same
         # because dynamic adjust is False:
         ENGINE._do_process_key_event(IBus.KEY_i, 0, 0)
         ENGINE._do_process_key_event(IBus.KEY_r, 0, 0)
         ENGINE._do_process_key_event(IBus.KEY_p, 0, 0)
-        self.assertEqual('怠', ENGINE.mock_preedit_text)
+        self.assertEqual('感', ENGINE.mock_preedit_text)
         self.assertEqual(
-            ['怠  1000 0',
-             '感  1000 0'],
+            ['感  1000 0',
+             '怠  900 0'],
             ENGINE._lookup_table.mock_candidates)
         ENGINE._do_process_key_event(IBus.KEY_2, 0, 0)
         self.assertEqual(
             [],
             ENGINE._lookup_table.mock_candidates)
         self.assertEqual('', ENGINE.mock_preedit_text)
-        self.assertEqual('感感', ENGINE.mock_committed_text)
+        self.assertEqual('怠怠', ENGINE.mock_committed_text)
         # Now set dynamic adjust to True:
         ENGINE.set_dynamic_adjust(True)
         # repeat the same input, result should still be the same
@@ -1334,74 +1397,74 @@ class Cangjie5TestCase(unittest.TestCase):
         ENGINE._do_process_key_event(IBus.KEY_i, 0, 0)
         ENGINE._do_process_key_event(IBus.KEY_r, 0, 0)
         ENGINE._do_process_key_event(IBus.KEY_p, 0, 0)
-        self.assertEqual('怠', ENGINE.mock_preedit_text)
+        self.assertEqual('感', ENGINE.mock_preedit_text)
         self.assertEqual(
-            ['怠  1000 0',
-             '感  1000 0'],
+            ['感  1000 0',
+             '怠  900 0'],
             ENGINE._lookup_table.mock_candidates)
         ENGINE._do_process_key_event(IBus.KEY_2, 0, 0)
         self.assertEqual(
             [],
             ENGINE._lookup_table.mock_candidates)
         self.assertEqual('', ENGINE.mock_preedit_text)
-        self.assertEqual('感感感', ENGINE.mock_committed_text)
-        # Try once again, now 感 should be preferred because
-        # 感 has been selected once while dynamic adjustment was used
-        # whereas 怠 has not been selected:
+        self.assertEqual('怠怠怠', ENGINE.mock_committed_text)
+        # Try once again, now 怠 should be preferred because
+        # 怠 has been selected once while dynamic adjustment was used
+        # whereas 感 has not been selected:
         ENGINE._do_process_key_event(IBus.KEY_i, 0, 0)
         ENGINE._do_process_key_event(IBus.KEY_r, 0, 0)
         ENGINE._do_process_key_event(IBus.KEY_p, 0, 0)
-        self.assertEqual('感', ENGINE.mock_preedit_text)
+        self.assertEqual('怠', ENGINE.mock_preedit_text)
         self.assertEqual(
-            ['感  1000 1',
-             '怠  1000 0'],
+            ['怠  900 1',
+             '感  1000 0'],
             ENGINE._lookup_table.mock_candidates)
         ENGINE._do_process_key_event(IBus.KEY_1, 0, 0)
         self.assertEqual(
             [],
             ENGINE._lookup_table.mock_candidates)
         self.assertEqual('', ENGINE.mock_preedit_text)
-        self.assertEqual('感感感感', ENGINE.mock_committed_text)
-        # Try once again with dynamic adjust, now 感 should have
+        self.assertEqual('怠怠怠怠', ENGINE.mock_committed_text)
+        # Try once again with dynamic adjust, now 怠 should have
         # user count 2:
         ENGINE._do_process_key_event(IBus.KEY_i, 0, 0)
         ENGINE._do_process_key_event(IBus.KEY_r, 0, 0)
         ENGINE._do_process_key_event(IBus.KEY_p, 0, 0)
-        self.assertEqual('感', ENGINE.mock_preedit_text)
+        self.assertEqual('怠', ENGINE.mock_preedit_text)
         self.assertEqual(
-            ['感  1000 2',
-             '怠  1000 0'],
+            ['怠  900 2',
+             '感  1000 0'],
             ENGINE._lookup_table.mock_candidates)
         ENGINE._do_process_key_event(IBus.KEY_1, 0, 0)
         self.assertEqual(
             [],
             ENGINE._lookup_table.mock_candidates)
         self.assertEqual('', ENGINE.mock_preedit_text)
-        self.assertEqual('感感感感感', ENGINE.mock_committed_text)
-        # 感 should have user count 3 now. This should be remembered
+        self.assertEqual('怠怠怠怠怠', ENGINE.mock_committed_text)
+        # 怠 should have user count 3 now. This should be remembered
         # in the user data base, i.e. if we switch off dynamic adjust
         # and later switch it back on, the previous user count should
         # still be there.
         #
         # Switch dynamic adjust off again:
         ENGINE.set_dynamic_adjust(False)
-        # And try again, now 怠 should be preferred again
-        # because of its smaller Big5 code and the user counts
+        # And try again, now 感 should be preferred again
+        # because of its higher weight in the cangjie5 table and the user counts
         # displayed should be 0:
         ENGINE._do_process_key_event(IBus.KEY_i, 0, 0)
         ENGINE._do_process_key_event(IBus.KEY_r, 0, 0)
         ENGINE._do_process_key_event(IBus.KEY_p, 0, 0)
-        self.assertEqual('怠', ENGINE.mock_preedit_text)
+        self.assertEqual('感', ENGINE.mock_preedit_text)
         self.assertEqual(
-            ['怠  1000 0',
-             '感  1000 0'],
+            ['感  1000 0',
+             '怠  900 0'],
             ENGINE._lookup_table.mock_candidates)
         ENGINE._do_process_key_event(IBus.KEY_2, 0, 0)
         self.assertEqual(
             [],
             ENGINE._lookup_table.mock_candidates)
         self.assertEqual('', ENGINE.mock_preedit_text)
-        self.assertEqual('感感感感感感', ENGINE.mock_committed_text)
+        self.assertEqual('怠怠怠怠怠怠', ENGINE.mock_committed_text)
         # Switch dynamic adjust on again:
         ENGINE.set_dynamic_adjust(True)
         # Try once again with dynamic adjust, now 感 should be preferred
@@ -1409,17 +1472,17 @@ class Cangjie5TestCase(unittest.TestCase):
         ENGINE._do_process_key_event(IBus.KEY_i, 0, 0)
         ENGINE._do_process_key_event(IBus.KEY_r, 0, 0)
         ENGINE._do_process_key_event(IBus.KEY_p, 0, 0)
-        self.assertEqual('感', ENGINE.mock_preedit_text)
+        self.assertEqual('怠', ENGINE.mock_preedit_text)
         self.assertEqual(
-            ['感  1000 3',
-             '怠  1000 0'],
+            ['怠  900 3',
+             '感  1000 0'],
             ENGINE._lookup_table.mock_candidates)
         ENGINE._do_process_key_event(IBus.KEY_1, 0, 0)
         self.assertEqual(
             [],
             ENGINE._lookup_table.mock_candidates)
         self.assertEqual('', ENGINE.mock_preedit_text)
-        self.assertEqual('感感感感感感感', ENGINE.mock_committed_text)
+        self.assertEqual('怠怠怠怠怠怠怠', ENGINE.mock_committed_text)
 
 class IpaXSampaTestCase(unittest.TestCase):
     def setUp(self):
